//
// progbar.c: the Progress Bar Control module.
//
// Copyright (C) 1999, 2000, Wei Yongming.
//
// Current maintainer: Wei Yonming.
//
// TODO: get rid of "double" floating point and ldiv() calls

/*
**  This library is free software; you can redistribute it and/or
**  modify it under the terms of the GNU Library General Public
**  License as published by the Free Software Foundation; either
**  version 2 of the License, or (at your option) any later version.
**
**  This library is distributed in the hope that it will be useful,
**  but WITHOUT ANY WARRANTY; without even the implied warranty of
**  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
**  Library General Public License for more details.
**
**  You should have received a copy of the GNU Library General Public
**  License along with this library; if not, write to the Free
**  Software Foundation, Inc., 59 Temple Place - Suite 330, Boston,
**  MA 02111-1307, USA
*/

/*
**  Alternatively, the contents of this file may be used under the terms 
**  of the Mozilla Public License (the "MPL License") in which case the
**  provisions of the MPL License are applicable instead of those above.
*/

//
// Note:
//   Originally by Zhao Jianghua. 
//
// Create date: 1999/8/29
//
// Modify records:
//
//  Who             When        Where       For What                Status
//-----------------------------------------------------------------------------
//  WEI Yongming    1999/10/27  Tsinghua    unsigned int            Finished
//  WEI Yongming    1999/10/27  Tsinghua    FPException fixing      Finished
//  WEI Yongming    2000/02/24  Tsinghua    Add MPL License         Finished
//  Kevin Tseng     2000/05/24  gv          port to microwin        ported
//

#include <stdio.h>
#include <stdlib.h>
#include <string.h>

#define MWINCLUDECOLORS
#include "windows.h"	/* windef.h, winuser.h */
#include "wintools.h" 	/* Draw3dBox */
#include "device.h" 	/* GdGetTextSize */

#define TEST	1	/* =1 for testing*/

typedef  struct tagPROGRESSDATA
{
    unsigned int nMin;
    unsigned int nMax;
    unsigned int nPos;
    unsigned int nStepInc;
}PROGRESSDATA;
typedef PROGRESSDATA* PPROGRESSDATA;

static LRESULT CALLBACK ProgressBarCtrlProc (HWND hwnd, UINT message, WPARAM wParam, LPARAM lParam);

#define  WIDTH_PBAR_BORDER  2 

int WINAPI MwRegisterProgressBarControl(HINSTANCE hInstance)
{
	WNDCLASS	wc;

	wc.style	= CS_HREDRAW | CS_VREDRAW | CS_DBLCLKS | CS_GLOBALCLASS;
	wc.lpfnWndProc	= (WNDPROC)ProgressBarCtrlProc;
	wc.cbClsExtra	= 0;
	wc.cbWndExtra	= 0;
	wc.hInstance	= hInstance;
	wc.hIcon	= NULL;
	wc.hCursor	= 0; /*LoadCursor(NULL, IDC_ARROW);*/
	wc.hbrBackground= GetStockObject(LTGRAY_BRUSH);
	wc.lpszMenuName	= NULL;
	wc.lpszClassName= "PROGBAR";

	return RegisterClass(&wc);
}

void pbarOnDraw (HWND hwnd, HDC hdc, 
            PROGRESSDATA* pData, BOOL fVertical, BOOL fErase)
{
    RECT    rcClient,rc;
    int     x, y, w, h, cw, ch, cb, x2;
    ldiv_t   ndiv_progress;
    unsigned int     nAllPart;
    unsigned int     nNowPart;
    int     whOne, nRem;
    int     ix, iy;
    int     i;
    char    szText[8] = "\0";
    int     step;

	HBRUSH hbr,hbrBlue;
	int maxw,charh;
    
    if (pData->nMax == pData->nMin)
        return;
    
    if ((pData->nMax - pData->nMin) > 5)
        step = 5;
    else
        step = 1;

    GetClientRect (hwnd, &rcClient);

    x = rcClient.left + WIDTH_PBAR_BORDER;
    y = rcClient.top + WIDTH_PBAR_BORDER;
#define RECTW(rc) (rc.right-rc.left)
#define RECTH(rc) (rc.bottom-rc.top)
    w = RECTW (rcClient) - (WIDTH_PBAR_BORDER << 1);
    h = RECTH (rcClient) - (WIDTH_PBAR_BORDER << 1);

    ndiv_progress = ldiv (pData->nMax - pData->nMin, step);
    nAllPart = ndiv_progress.quot;
    
    ndiv_progress = ldiv (pData->nPos - pData->nMin, step);
    nNowPart = ndiv_progress.quot;
    if (fVertical)
        ndiv_progress = ldiv (h, nAllPart);
    else
        ndiv_progress = ldiv (w, nAllPart);
        
    whOne = ndiv_progress.quot;
    nRem = ndiv_progress.rem;

    if (fErase) {
#if 0
//-     SetBrushColor (hdc, COLOR_darkgray);
//-     FillBox (hdc, x, y, w, h);
#else
	rc.left=x; rc.top=y; rc.bottom=y+h; rc.right=x+w;
	//jmprintf("1 rc(l,t,r,b)=%d,%d,%d,%d\n",rc.left,rc.top,rc.right,rc.bottom);
	FillRect(hdc,&rc,GetStockObject(DKGRAY_BRUSH));
#endif
    }

#if 0
//- SetBrushColor (hdc, COLOR_blue);
#else
    hbrBlue=CreateSolidBrush(BLUE);
#endif

    if(whOne >= 4) {
        if (fVertical) {
            for (i = 0, iy = y + h - 1; i < nNowPart; ++i) {
                if ((iy - whOne) < y) 
                    whOne = iy - y;

#if 0
//-             FillBox (hdc, x, iy - whOne, w, whOne - 1);
#else
		rc.left=x; rc.top=iy-whOne;rc.right=x+w; rc.bottom=rc.top+whOne-1;
		//jmprintf("2 rc(l,t,r,b)=%d,%d,%d,%d\n",rc.left,rc.top,rc.right,rc.bottom);
		FillRect(hdc,&rc,hbrBlue);
#endif

                iy -= whOne;
                if(nRem > 0) {
                    iy --;
                    nRem --;
                }
            }
        }
        else {
            for (i = 0, ix = x + 1; i < nNowPart; ++i) {
                if ((ix + whOne) > (x + w)) 
                    whOne = x + w - ix;
#if 0
//-             FillBox (hdc, ix, y, whOne - 1, h);
#else
		rc.left=ix; rc.top=y;rc.right=ix+whOne-1; rc.bottom=y+h;
		//jmprintf("3 rc(l,t,r,b)=%d,%d,%d,%d\n",rc.left,rc.top,rc.right,rc.bottom);
		FillRect(hdc,&rc,hbrBlue);
#endif
                ix += whOne;
                if(nRem > 0) {
                    ix ++;
                    nRem --;
                }
            }
        }
    }
    else {
        // no vertical support
        double d = (nNowPart*1.0)/nAllPart;
#if 0        
//-     int maxw = GetSysCharWidth () << 2;
//-     int charh = GetSysCharHeight ();
#else
	GdGetTextSize(hdc->font->pfont,"X",1,
		&cw,&ch,&cb,MWTF_ASCII);
	maxw=cw << 2;
	charh=ch;
#endif
        if ( (int)(d*100) > 50 )
	{
#if 0
            //- SetBrushColor (hdc, COLOR_blue);
#else
		hbr=hbrBlue;
#endif
	}
        else
	{
#if 0
            //- SetBrushColor (hdc, COLOR_darkgray);
#else
		hbr=GetStockObject(DKGRAY_BRUSH);
#endif
	}
#if 0
//-     FillBox (hdc, x + ((w - maxw)>>1), y + ((h - charh) > 1), maxw, charh);
#else
	rc.left=x+((w-maxw)>>1); rc.top=y+((h-charh) > 1);rc.right=rc.left+maxw; rc.bottom=rc.top+charh;
	//jmprintf("4: rc(l,t,r,b)=%d,%d,%d,%d\n",rc.left,rc.top,rc.right,rc.bottom);
	FillRect(hdc,&rc,hbr);
#endif
#if 0
//-     SetBrushColor (hdc, COLOR_blue);
//-     FillBox (hdc, x, y, (int)(w*d), h - 1);
#else
	x2=(int)(w*d);
//pbar: fblin8.c:88: linear8_drawhorzline: Assertion `x2 >= x1' failed.
//jmt: FillRect fail(bug?) if (x2==x1)
	if (x2)
	{
		rc.left=x; rc.top=y; rc.right=x+x2; rc.bottom=y+h-1;
		//jmprintf("5: rc(l,t,r,b)=%d,%d,%d,%d\n",rc.left,rc.top,rc.right,rc.bottom);
		FillRect(hdc,&rc,hbrBlue);
	}
#endif
        SetTextColor (hdc, LTGRAY/*COLOR_lightwhite*/);
        SetBkMode (hdc, TRANSPARENT/*BM_TRANSPARENT*/);
        sprintf (szText, "%d%%", (int)(d*100));
#if 0
//-     TextOut (hdc, x + ((w - GetSysCharWidth () * strlen (szText) )>>1), 
//-                   y + ((h - GetSysCharHeight() )>>1), 
//-                   szText);
#else
	GdGetTextSize(hdc->font->pfont,szText,strlen(szText),
		&cw,&ch,&cb,MWTF_ASCII);
        TextOut (hdc, x + ((w - cw )>>1), 
                      y + ((h - ch )>>1), 
                      szText, strlen(szText));
#endif
    }

    DeleteObject(hbrBlue);
}

static void pbarNormalizeParams (const HWND pCtrl, 
                PROGRESSDATA* pData, BOOL fNotify)
{
    if (pData->nPos > pData->nMax) {
        if (fNotify)
            SendMessage (GetParent ((HWND)pCtrl), WM_COMMAND, 
		(WPARAM)MAKELONG (pCtrl->id, PBN_REACHMAX), (LPARAM)pCtrl);
        pData->nPos = pData->nMax;
    }

    if (pData->nPos < pData->nMin) {
        if (fNotify)
            SendMessage (GetParent ((HWND)pCtrl), WM_COMMAND, 
		(WPARAM)MAKELONG (pCtrl->id, PBN_REACHMIN), (LPARAM)pCtrl);
        pData->nPos = pData->nMin;
    }
}

LRESULT CALLBACK
ProgressBarCtrlProc (HWND hwnd, UINT message, WPARAM wParam, LPARAM lParam)
{
    HDC           hdc;
    HWND	  pCtrl;
    PROGRESSDATA* pData;
    BOOL          fErase;
    RECT	rc;
    PAINTSTRUCT ps;
    int		pos;
    
    pCtrl = hwnd;
    
    switch (message)
    {
        case WM_CREATE:
            if (!(pData = malloc (sizeof (PROGRESSDATA)))) {
                fprintf(stderr, "Create progress bar control failure!\n");
                return -1;
            }
            
            pData->nMax     = 100;
            pData->nMin     = 0;
#if TEST
            pData->nPos     = 50;
#else
            pData->nPos     = 0;
#endif
            pData->nStepInc = 10;

            pCtrl->userdata = (DWORD)pData;
        break;
    
        case WM_DESTROY:
            free ((void *)(pCtrl->userdata));
        break;

        case WM_GETDLGCODE:
            return DLGC_STATIC;

        case WM_GETTEXTLENGTH:
        case WM_GETTEXT:
        case WM_SETTEXT:
            return -1;

	case WM_NCPAINT:
		if(GetWindowLong(hwnd, GWL_STYLE) & WS_BORDER) {
			GetWindowRect(hwnd, &rc);
			hdc = GetWindowDC(hwnd);
			Draw3dBox(hdc, rc.left, rc.top, rc.right-rc.left,
				rc.bottom-rc.top, GetSysColor(COLOR_BTNSHADOW),
				GetSysColor(COLOR_BTNHIGHLIGHT));
		}
	    break;

	case WM_PAINT:
            hdc = BeginPaint (hwnd,&ps);
            pbarOnDraw (hwnd, hdc, (PROGRESSDATA *)pCtrl->userdata, 
                            hwnd->style & PBS_VERTICAL, TRUE);
            EndPaint (hwnd, &ps);
            break;

        case PBM_SETRANGE:
            pData = (PROGRESSDATA *)pCtrl->userdata;
            pData->nMin = min (wParam, lParam);
            pData->nMax = max (wParam, lParam);
            if (pData->nPos > pData->nMax)
                pData->nPos = pData->nMax;
            if (pData->nPos < pData->nMin)
                pData->nPos = pData->nMin;
        break;
        
        case PBM_SETSTEP:
            pData = (PROGRESSDATA *)pCtrl->userdata;
            pData->nStepInc = wParam;
        break;
        
        case PBM_SETPOS:
            pData = (PROGRESSDATA *)pCtrl->userdata;
            
            if (pData->nPos == wParam)
                break;

            fErase = (wParam < pData->nPos);
            pData->nPos = wParam;
            pbarNormalizeParams (pCtrl, pData, hwnd->style & PBS_NOTIFY);
	    InvalidateRect(hwnd, NULL, fErase);
        break;
        
        case PBM_DELTAPOS:
            pData = (PROGRESSDATA *)pCtrl->userdata;

            if (wParam == 0)
                break;
            
            fErase = (wParam < 0);
            pData->nPos += wParam;
            pbarNormalizeParams (pCtrl, pData, hwnd->style & PBS_NOTIFY);
	    InvalidateRect(hwnd, NULL, fErase);
        break;
        
        case PBM_STEPIT:
            pData = (PROGRESSDATA *)pCtrl->userdata;
            
            if (pData->nStepInc == 0)
                break;

            fErase = (pData->nStepInc < 0);
            pData->nPos += pData->nStepInc;
            pbarNormalizeParams (pCtrl, pData, hwnd->style & PBS_NOTIFY);
	    InvalidateRect(hwnd, NULL, fErase);
        break;
            
#if TEST
	case WM_LBUTTONDOWN:
	case WM_LBUTTONDBLCLK:
                pData = (PROGRESSDATA *)pCtrl->userdata;
		pos = pData->nPos;
		pos += 10;
		if (pos > 100)
			pos = 0;
		SendMessage(hwnd, PBM_SETPOS, pos, 0L);
		break;
#endif
        default:
    		return DefWindowProc (hwnd, message, wParam, lParam);
    }
    return 0;
}
